
.red {
  color:#990000
}

#wrap {
  min-height: 100%;
  height: auto;
  margin: 0 auto -6ex;
  padding: 0 0 6ex;
}

.pattern_tab {
  margin: 1ex;
}

.pattern_brat {
  margin-top: 1ex;
}

.label {
  color: #777777;
  font-size: small;
}

.footer {
  bottom: 0;
  width: 100%;
  /* Set the fixed height of the footer here */
  height: 5ex;
  padding-top: 1ex;
  margin-top: 1ex;
  background-color: #f5f5f5;
}

.corenlp_error {
  margin-top: 2ex;
}

/* Styling for parse graph */
.node rect {
  stroke: #333;
  fill: #fff;
}

.parse-RULE rect {
  fill: #C0D9AF;
}

.parse-TERMINAL rect {
  stroke: #333;
  fill: #EEE8AA;
}

.node.highlighted {
  stroke: #ffff00;
}

.edgePath path {
  stroke: #333;
  fill: #333;
  stroke-width: 1.5px;
}

.parse-EDGE path {
  stroke: DarkGray;
  fill: DarkGray;
  stroke-width: 1.5px;
}

.logo {
    font-family: "Lato", "Gill Sans MT", "Gill Sans", "Helvetica", "Arial", sans-serif;
    font-style: italic;
}
